/*-
 * ~~~~~~licensing~~~~~~
 * generator-api
 * ==========
 * Copyright (C) 2020 - 2021 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.base.generator.template.impl;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.Optional;
import java.util.Properties;
import java.util.stream.Collectors;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.ServiceRegistration;
import org.osgi.util.tracker.BundleTrackerCustomizer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;

import ru.entaxy.base.generator.template.Template;
import ru.entaxy.base.generator.template.TemplateImpl;
import ru.entaxy.base.generator.template.TemplateService;

public class TemplateBundleTrackerCustomizer implements BundleTrackerCustomizer<TemplateProvider> {

	private static final Logger log = LoggerFactory.getLogger(TemplateBundleTrackerCustomizer.class);
	
	@Override
	public TemplateProvider addingBundle(Bundle bundle, BundleEvent event) {
		try {
			TemplateProvider provider = createTemplateProvider(bundle);
			if (provider != null)
				register(provider);
			return provider;
		} catch (Exception e) {
			log.error("Error creating TemplateProvider for bundle ["
					+ bundle.getBundleId() + "]"
					, e);
		}
		return null;
	}

	@Override
	public void modifiedBundle(Bundle bundle, BundleEvent event, TemplateProvider object) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void removedBundle(Bundle bundle, BundleEvent event, TemplateProvider object) {
		
		unregister(object);
		
	}

	protected TemplateProvider createTemplateProvider(Bundle bundle) throws Exception {
		
		TemplateProvider result = new TemplateProvider();
		result.bundle = bundle;
		result.bundleContext = bundle.getBundleContext();
		
		boolean isTemplateProvider = Boolean.valueOf(
				Optional
				.ofNullable(bundle.getHeaders().get(TemplateService.TEMPLATE_PROVIDER_HEADER))
				.orElse("false")
			);
		if (!isTemplateProvider)
			return null;
		
		URL metadataUrl = bundle.getEntry(TemplateProvider.TEMPLATE_PATH + "/metadata.json");
		String metadata = new BufferedReader (
				  new InputStreamReader(
						  metadataUrl.openStream(), StandardCharsets.UTF_8))
					.lines()
					.collect(Collectors.joining("\n"));
		
		GsonBuilder builder = new GsonBuilder();
		Gson gson = builder.create();

		JsonElement je = (new JsonParser()).parse(metadata);
		JsonObject root = je.getAsJsonObject();

		JsonElement templatesElement = root.get("templates");
		if (templatesElement == null)
			return null;
		if (!templatesElement.isJsonArray())
			return null;
		JsonArray templates = templatesElement.getAsJsonArray();
		for (int i=0; i<templates.size(); i++) {
			TemplateMetadata meta = gson.fromJson(templates.get(i), TemplateMetadata.class);
			ProvidedTemplate providedTemplate = new ProvidedTemplate(meta);
			result.providedTemplates.add(providedTemplate);
		}
		
		return result;
	}
	
	protected void register(TemplateProvider templateProvider) {
		
		for (ProvidedTemplate provided: templateProvider.providedTemplates) {
			
			TemplateImpl template = new TemplateImpl();
			template.setBundleContext(templateProvider.bundleContext);
			template.load(provided);
			// TODO fill object
			
			Dictionary<String, String> props = new Hashtable<String, String>();
			props.put(TemplateService.PROP_ID, provided.getId());
			props.put(TemplateService.PROP_NAME, provided.getName());
			props.put(TemplateService.PROP_TYPE, provided.getType());
			props.put(TemplateService.PROP_DESCRIPTION, provided.getDescription());
			
			ServiceRegistration<Template> registration = 
					templateProvider.bundleContext.registerService(
							Template.class, template, props
					);
			
			provided.registration = registration;
		}
		
	}

	protected void unregister(TemplateProvider templateProvider) {
		
		for (ProvidedTemplate provided: templateProvider.providedTemplates) {
			provided.registration.unregister();
		}
		
	}
	
}
