/*-
 * ~~~~~~licensing~~~~~~
 * profile-management
 * ==========
 * Copyright (C) 2020 - 2021 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.esb.platform.base.management.core.utils;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import org.osgi.framework.BundleContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.entaxy.esb.platform.base.management.core.ManagementCore;
import ru.entaxy.platform.base.support.osgi.tracker.BundleTrackerCustomizerListener;

public abstract class ManagedObjectsListener<T, S, I, C extends I> implements BundleTrackerCustomizerListener<T> {

	private static final Logger log = LoggerFactory.getLogger(ManagedObjectsListener.class);
	
	protected Map<String, ManagedObjectDescriptor<S, I, C>> managed = new HashMap<>();
	
	protected BundleContext bundleContext;
	
	protected Class<I> serviceClass;
	
	public ManagedObjectsListener(BundleContext bundleContext, Class<I> serviceClass) {
		super();
		this.bundleContext = bundleContext;
		this.serviceClass = serviceClass;
	}
	
	public Map<String, ManagedObjectDescriptor<S, I, C>> getManaged() {
		return managed;
	}

	@Override
	public void added(T managedObject) {
		Iterator<S> iter = getIterator(managedObject);
		while (iter.hasNext()) {
			S object = iter.next();
			String key = getKeyValue(object);
			if (managed.containsKey(key)) {
				managed.get(key).registration.unregister();
			}
			C service = createService(object);
			if (service == null)
				continue;
			ManagedObjectDescriptor<S, I, C> descriptor = new ManagedObjectDescriptor<>();
			descriptor.object = object;
			descriptor.service = service;
			Hashtable props = new Hashtable<>();
			props.put(ManagementCore.JMX_OBJECTNAME, getJmxObjectName(object));
			descriptor.registration = bundleContext.registerService(this.serviceClass, service, props);
			managed.put(key, descriptor);
		}
	}

	@Override
	public void removed(T managedObject) {
		Iterator<S> iter = getIterator(managedObject);
		while (iter.hasNext()) {
			S object = iter.next();
			String key = getKeyValue(object);
			if (managed.containsKey(key)) {
				managed.get(key).registration.unregister();
				managed.remove(key);
			}
		}
	}

	@Override
	public void modified(T managedObject) {
		log.info(">> modified");
		Iterator<S> iter = getIterator(managedObject);
		while (iter.hasNext()) {
			S object = iter.next();
			String key = getKeyValue(object);
			if (managed.containsKey(key)) {
				C service = managed.get(key).service;
				if (service instanceof BundleAwareMBean) {
					((BundleAwareMBean)service).refreshBundleInfo();
					log.info(">> refreshBundleInfo");
				} else {
					log.info(">> NOT BundleAwareMBean");
				}
			}
		}
	}
	
	protected abstract Iterator<S> getIterator(T managedObject);
	
	protected abstract String getKeyValue(S object);
	
	protected abstract C createService(S managedObject);
	
	protected abstract String getJmxObjectName(S managedObject);
}
