/*-
 * ~~~~~~licensing~~~~~~
 * artifact-management
 * ==========
 * Copyright (C) 2020 - 2021 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.core.artifact;

import java.util.Calendar;

import org.apache.camel.tooling.model.Strings;

import ru.entaxy.platform.base.support.CommonUtils;

public class ArtifactCoordinates {

	public static final ArtifactCoordinates EMPTY = new ArtifactCoordinates(); 
	
	protected String groupId;
	protected String artifactId;
	protected String version = "1.0";
	protected String type;
	protected String classifier;

	protected String versionQualifier = "";
	
	public ArtifactCoordinates groupId(String groupId) {
		setGroupId(groupId);
		return this;
	}
	
	public ArtifactCoordinates artifactId(String artifactId) {
		setArtifactId(artifactId);
		return this;
	}
	
	public ArtifactCoordinates version(String version) {
		this.versionQualifier = "";
		setVersion(version);
		return this;
	}
	
	public ArtifactCoordinates timestampedVersion(String version) {
		version(version);
		this.versionQualifier = "-" + Calendar.getInstance().getTimeInMillis();
		return this;
	}
	
	public ArtifactCoordinates timestamped() {
		return qualifier(Calendar.getInstance().getTimeInMillis()+"");
	}
	
	public ArtifactCoordinates timestamped(String timestamp) {
		if (CommonUtils.isValid(timestamp))
			return qualifier(timestamp);
		else
			return timestamped();
	}
	
	public ArtifactCoordinates type(String type) {
		setType(type);
		return this;
	}
	
	public ArtifactCoordinates qualifier(String qualifier) {
		if (CommonUtils.isValid(qualifier))
			this.versionQualifier = "-" + qualifier;
		else
			this.versionQualifier = "";
		return this;
	}
	
	public ArtifactCoordinates classifier(String classifier) {
		setClassifier(classifier);
		return this;
	}
	
	public ArtifactCoordinates set(ArtifactCoordinates newCoordinates) {
		groupId(newCoordinates.getGroupId())
		.artifactId(newCoordinates.getArtifactId())
		.version(newCoordinates.getVersion())
		.type(newCoordinates.getType())
		.classifier(newCoordinates.getClassifier());
		return this;
	}
	
	public ArtifactCoordinates update(ArtifactCoordinates newCoordinates) {
		if (null != newCoordinates.getGroupId())
			groupId(newCoordinates.getGroupId());
		if (null != newCoordinates.getArtifactId())
			artifactId(newCoordinates.getArtifactId());
		if (null != newCoordinates.getVersion())
			version(newCoordinates.getVersion());
		if (null != newCoordinates.getType())
			type(newCoordinates.getType());
		if (null != newCoordinates.getClassifier())
			classifier(newCoordinates.getClassifier());
		return this;
	}
	
	public boolean isSnapshot() {
		return !Strings.isNullOrEmpty(this.version)
			&& this.version.endsWith("-SNAPSHOT");
	}
	
	public String getGroupId() {
		return groupId;
	}
	public void setGroupId(String groupId) {
		this.groupId = groupId;
	}
	public String getArtifactId() {
		return artifactId;
	}
	public void setArtifactId(String artifactId) {
		this.artifactId = artifactId;
	}
	public String getVersion() {
		return version + versionQualifier;
	}
	public void setVersion(String version) {
		if (!CommonUtils.isValid(version))
			this.version = "1.0.0";
		else {
			String[] splitted = version.split("\\.");
			if (splitted.length == 2)
				this.version = version + ".0";
			else if ((splitted.length==3) && (!CommonUtils.isValid(splitted[2])))
				this.version = version + "0";
			else this.version = version;
		}
	}
	public String getType() {
		return type;
	}
	public void setType(String type) {
		this.type = type;
	}
	public String getClassifier() {
		return classifier;
	}
	public void setClassifier(String classifier) {
		this.classifier = classifier;
	}
	
	@Override
	public String toString() {
		String result = String.format("%s/%s/%s%s", groupId, artifactId, version, versionQualifier);
		if (!Strings.isNullOrEmpty(type)) {
			result += "/" + type;
			if (!Strings.isNullOrEmpty(classifier))
				result += "/" + classifier;
		}
		return result;
	}
	
}
