/*-
 * ~~~~~~licensing~~~~~~
 * artifact-management
 * ==========
 * Copyright (C) 2020 - 2021 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.core.artifact;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Enumeration;
import java.util.Properties;
import java.util.jar.Manifest;

import org.osgi.framework.Constants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import ru.entaxy.platform.base.support.CommonUtils;
import ru.entaxy.platform.base.support.xml.CommonXMLUtils;
import ru.entaxy.platform.core.artifact.annotation.ArtifactSupport;

@ArtifactSupport(supportedCategory = Artifact.ARTIFACT_CATEGORY_BLUEPRINT
	, supportedContentClasses = {String.class, Document.class, byte[].class}
	, defaultArtifactClassifier = Artifact.ARTIFACT_CATEGORY_BLUEPRINT
	, defaultArtifactType = "xml")
public class Blueprint extends DefaultArtifact {

	private final static Logger log = LoggerFactory.getLogger(Blueprint.class);
	
	@Override
	public void setContent(Object content) {
		super.setContent(content);
		if (this.content.getClass().equals(byte[].class)) {
			String s = new String((byte[])this.content, StandardCharsets.UTF_8);
			this.content = s;
		}
		if (this.content.getClass().equals(String.class)) {
			try {
				Document d = CommonXMLUtils.parseString(false, (String)this.content);
				this.content = d;
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}
	
	@Override
	public void toFile(File destination) {
		if (this.content instanceof String)
			super.toFile(destination);
		else if (this.content instanceof Document) {
			Document doc = (Document)this.content;
			try {
				prepareManifest(doc);
				CommonXMLUtils.saveDocument(doc, destination);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		else if (this.content instanceof byte[]) {
			// TODO to be implemented
		}
		
	}
	
	@Override
	public byte[] asByteArray() {
		if (this.content instanceof byte[])
			return (byte[])this.content;
		if (this.content instanceof String)
			return ((String)this.content).getBytes();
		if (this.content instanceof Document)
			try {
				Document doc = (Document)this.content;
				prepareManifest(doc);
				return CommonXMLUtils.doc2string(doc).getBytes();
			} catch (Exception e) {
				e.printStackTrace();
			}
		return new byte[] {};
	}
	
	protected void prepareManifest(Document doc) {
		Manifest m = new Manifest();
		m.getMainAttributes().putValue("Manifest-Version", "2");
	    m.getMainAttributes().putValue(Constants.BUNDLE_MANIFESTVERSION, "2");
	    m.getMainAttributes().putValue(Constants.BUNDLE_SYMBOLICNAME
	    		, this.coordinates.getGroupId() + "." + this.coordinates.getArtifactId());
	    m.getMainAttributes().putValue(Constants.BUNDLE_NAME
	    		, this.coordinates.getGroupId() + "." + this.coordinates.getArtifactId());
        m.getMainAttributes().putValue(Constants.BUNDLE_VERSION, this.coordinates.getVersion().replaceAll("-", "."));
        m.getMainAttributes().putValue(Constants.DYNAMICIMPORT_PACKAGE, "*");
        
        m.getMainAttributes().putValue(Artifact.MANIFEST_HEADER_ENTAXY_BUNLDE, "true");
        m.getMainAttributes().putValue(Artifact.MANIFEST_HEADER_ENTAXY_RUNTIME, "true");
        
		NodeList l = doc.getElementsByTagName("manifest");
        if (l != null) {
            for (int i = 0; i < l.getLength(); i++) {
                Element e = (Element) l.item(i);
                String text = e.getTextContent();
                Properties props = new Properties();
                try {
					props.load(new ByteArrayInputStream(text.trim().getBytes()));
				} catch (IOException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
                Enumeration en = props.propertyNames();
                while (en.hasMoreElements()) {
                    String k = (String) en.nextElement();
                    String v = props.getProperty(k);
                    m.getMainAttributes().putValue(k, v);
                }
                e.getParentNode().removeChild(e);
            }
        }
        
        // add capabilities info
        String requiredCapabilities = getRequiredCapabilitiesValue();
        String providedCapabilities = getProvidedCapabilitiesValue();
        
        if (CommonUtils.isValid(requiredCapabilities)) {
        	Object v = m.getMainAttributes().get(Constants.REQUIRE_CAPABILITY);
        	String value = (v==null?"":v.toString());
        	if (CommonUtils.isValid(value))
        		value += "," + requiredCapabilities;
        	else
        		value = requiredCapabilities;
        	m.getMainAttributes().putValue(Constants.REQUIRE_CAPABILITY, value);
        }
        
        if (CommonUtils.isValid(providedCapabilities)) {
        	Object v = m.getMainAttributes().get(Constants.PROVIDE_CAPABILITY);
        	String value = (v==null?"":v.toString());
        	if (CommonUtils.isValid(value))
        		value += "," + providedCapabilities;
        	else
        		value = providedCapabilities;
        	m.getMainAttributes().putValue(Constants.PROVIDE_CAPABILITY, value);
        }
        
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        String manifestContent = "";
        try {
			m.write(bos);
			manifestContent = bos.toString(StandardCharsets.ISO_8859_1);
			manifestContent = manifestContent.replaceAll("\r", "").replaceAll("\n ", "");
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
        Element manifest = doc.createElementNS("http://karaf.apache.org/xmlns/deployer/blueprint/v1.0.0", "manifest");
        manifest.setTextContent(manifestContent);
        doc.getDocumentElement().insertBefore(manifest, doc.getDocumentElement().getFirstChild());
        
        // copy manifest data to artifact properties
        for (Object key: m.getMainAttributes().keySet()) {
        	this.properties.put(key.toString(), m.getMainAttributes().get(key));
        }
	}
}
