/*-
 * ~~~~~~licensing~~~~~~
 * artifact-management
 * ==========
 * Copyright (C) 2020 - 2021 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.core.artifact.installer.builder.impl;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Map;

import org.apache.karaf.bundle.command.Install;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.entaxy.platform.core.artifact.DeployedArtifact;
import ru.entaxy.platform.core.artifact.installer.builder.InstallationResult;
import ru.entaxy.platform.core.artifact.installer.builder.Installer;
import ru.entaxy.platform.core.artifact.installer.builder.typed.BundleInstaller;
import ru.entaxy.platform.core.artifact.installer.builder.typed.TypedInstaller;
import ru.entaxy.platform.core.artifact.installer.builder.typed.impl.TypedInstallerHelper;
import ru.entaxy.platform.core.artifact.installer.builder.typed.impl.TypedInstallerImpl;
import ru.entaxy.platform.core.artifact.legacy.BundleController;

public abstract class AbstractInstaller<T extends Installer<T>, H extends TypedInstallerHelper> 
		implements BundleController {

	private static final Logger log = LoggerFactory.getLogger(AbstractInstaller.class);
	
	protected DeployedArtifact artifact = null;
	protected String sourceLocation = null;
	protected String bundleName = null;
	
	protected Map<String, Class<? extends TypedInstallerImpl>> typedInstallerClasses = new HashMap<>();
	protected Map<Class<? extends TypedInstallerImpl>, Class<? extends H>> typedHelpeClasses = new HashMap<>();

	
	// for BundleInstaller implementation
	protected BundleInstaller bundleInstaller = null;
	
	protected TypedInstallerImpl typedInstaller = null;

	protected AbstractInstaller() {
		this.typedInstallerClasses = new HashMap<>();
		this.typedHelpeClasses = new HashMap<>();
	}

	protected void doCreateTypedInstaller(Class<? extends TypedInstallerImpl> clazz) {
		Constructor<?> constructor;
		try {
			constructor = clazz.getConstructor();
			this.typedInstaller = (TypedInstallerImpl)constructor.newInstance();
		} catch (NoSuchMethodException | SecurityException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (InstantiationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IllegalArgumentException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	protected void createTypedInstaller() {
		Class<? extends TypedInstallerImpl> clazz =
				artifact==null
				?null
				:typedInstallerClasses.get(artifact.getArtifact().getCategory());
		if (clazz == null)
			// TODO throw exception
			return;
		doCreateTypedInstaller(clazz);
		if (this.typedInstaller != null) {
			initTypedInstaller();
		}
	}

	protected void findTypedInstaller(Class<?> type) {
		Class<? extends TypedInstallerImpl> clazz = null;
		for (Class<? extends TypedInstallerImpl> cl : typedInstallerClasses.values())
			if ((cl != null) && type.isAssignableFrom(cl)) {
				clazz = cl;
				break;
			}
		if (clazz == null)
			// TODO throw exception
			return;
		doCreateTypedInstaller(clazz);
		if (this.typedInstaller != null) {
			initTypedInstaller();
		}
	}
	
	protected H createHelper() {
		Class<? extends H> clazz = typedHelpeClasses.get(typedInstaller.getClass());
		if (clazz == null)
			return createDefaultHelper();
		Constructor<?> constructor;
		String errorMessage = "Error creating typed installer helper";
		try {
			constructor = clazz.getConstructor();
			return (H)constructor.newInstance();
		} catch (NoSuchMethodException | SecurityException e) {
			log.error(errorMessage, e);
		} catch (InstantiationException e) {
			log.error(errorMessage, e);
		} catch (IllegalAccessException e) {
			log.error(errorMessage, e);
		} catch (IllegalArgumentException e) {
			log.error(errorMessage, e);
		} catch (InvocationTargetException e) {
			log.error(errorMessage, e);
		}
	
		return createDefaultHelper();
	}
	
	protected abstract H createDefaultHelper(); 
	
	protected void refreshTypedInstaller() {
		this.typedInstaller.clearInstaller();
		this.typedInstaller.setArtifact(this.artifact);
		this.typedInstaller.setSourceLocation(this.sourceLocation);
		this.typedInstaller.setBundleName(this.bundleName);
	}
	
	protected void initTypedInstaller() {
		refreshTypedInstaller();
		
		H helper = createHelper();
		initHelper(helper);
		this.typedInstaller.setHelper(helper);
	}	

	protected void initHelper(H helper) {
		
	};
	
	public <S extends TypedInstaller> S typed(Class<S> type) {
		if (this.typedInstaller == null)
			createTypedInstaller();
		if (this.typedInstaller == null)
			findTypedInstaller(type);
		if (this.typedInstaller == null)
			// TODO throw exception
			return null;
		if (!type.isAssignableFrom(this.typedInstaller.getClass()))
			// TODO throw exception
			return null;
		return type.cast(this.typedInstaller);
	}

	public InstallationResult install() {
		if (typedInstaller == null) {
			createTypedInstaller();
		}
		if (typedInstaller != null)
			return typedInstaller.install();
		return null;
	}

	public InstallationResult uninstall() {
		if (typedInstaller == null) {
			createTypedInstaller();
		}
		if (typedInstaller != null)
			return typedInstaller.uninstall();
		return null;
	}

	public T artifact(DeployedArtifact artifact) {
		setArtifact(artifact);
		return (T)this;
	}

	public T sourceLocation(String sourceLocation) {
		setSourceLocation(sourceLocation);
		return (T)this;
	}
	
	public void setArtifact(DeployedArtifact artifact) {
		this.artifact = artifact;
	}

	public void setSourceLocation(String sourceLocation) {
		this.sourceLocation = sourceLocation;
	}

	public T bundleName(String bundleName) {
		setBundleName(bundleName);
		return (T) this;
	}
	
	public void setBundleName(String bundleName) {
		this.bundleName = bundleName;
	}

	public BundleController bundleController() {
		return (BundleController) this;
	} 

	protected BundleInstaller createBundleInstaller() {
		return typed(BundleInstaller.class);
	}
	
	protected void initBundleInstaller() {
		if (this.bundleInstaller == null)
			this.bundleInstaller = createBundleInstaller();
	}
	
	protected void clear() {
		this.artifact = null;
		this.sourceLocation = null;
		this.bundleName = null;
	}
	
	/*
	 * BundleController implementation
	 */

    public String installAndStartBundle(String bundleUrl, String bundleName) throws Exception {
    	initBundleInstaller();
    	clear();
    	this.sourceLocation(bundleUrl);
    	refreshTypedInstaller();
    	InstallationResult result = this.bundleInstaller.update(bundleName).installOnlyIfMissing().start().install();
    	return result.getProperties()
    			.getOrDefault(InstallationResult.PROP_BUNDLE_STATUS, "RESOLVED")
    			.toString();
    };

    public void installBundle(String bundleUrl) throws Exception {
    	initBundleInstaller();
    	clear();
    	this.sourceLocation(bundleUrl);
    	refreshTypedInstaller();
    	this.bundleInstaller.update().installOnlyIfMissing().install();
    };
    
    public String updateBundle(String bundleName) throws Exception {
    	initBundleInstaller();
    	clear();
    	InstallationResult result = this.bundleInstaller.update(bundleName).updateOnly().install();
    	return result.getProperties()
    			.getOrDefault(InstallationResult.PROP_BUNDLE_STATUS, "Resolved")
    			.toString();
    };

    public String startBundle(String bundleName) throws Exception {
    	initBundleInstaller();
    	clear();
    	bundleName(bundleName);
    	refreshTypedInstaller();
    	InstallationResult result = this.bundleInstaller.startOnly().install();
    	return result.getProperties().getOrDefault(InstallationResult.PROP_BUNDLE_STATUS
    			, "Resolved").toString();
    };

    public String stopBundle(String bundleName) throws Exception {
    	initBundleInstaller();
    	clear();
    	bundleName(bundleName);
    	refreshTypedInstaller();
    	InstallationResult result = this.bundleInstaller.stopOnly().install();
    	return result.getProperties().getOrDefault(InstallationResult.PROP_BUNDLE_STATUS
    			, "Resolved").toString();
    };

    public void uninstallBundle(String bundleName) throws Exception {
    	initBundleInstaller();
    	clear();
    	bundleName(bundleName);
    	refreshTypedInstaller();
    	this.bundleInstaller.uninstall();
    };	
	
}
