/*-
 * ~~~~~~licensing~~~~~~
 * artifact-management
 * ==========
 * Copyright (C) 2020 - 2021 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.core.artifact.repository.impl;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.util.Calendar;

import org.apache.camel.tooling.model.Strings;
import org.apache.karaf.cave.repository.Repository;
import org.apache.karaf.cave.repository.RepositoryService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.entaxy.platform.core.artifact.Artifact;
import ru.entaxy.platform.core.artifact.ArtifactCoordinates;
import ru.entaxy.platform.core.artifact.DeployedArtifact;
import ru.entaxy.platform.core.artifact.repository.ArtifactRepository;
import ru.entaxy.platform.core.artifact.repository.impl.remote.RemoteRepository;
import ru.entaxy.platform.core.artifact.repository.impl.remote.RemoteRepositoryDescriptor;
import ru.entaxy.platform.core.artifact.repository.impl.remote.RemoteRepositoryFactory;

public class ArtifactRepositoryImpl implements ArtifactRepository {
	
	private static final Logger log = LoggerFactory.getLogger(ArtifactRepositoryImpl.class);
	
	protected boolean isHealthy = false;
	
	protected RepositoryService repositoryService;
	
	protected RepositoryDescriptor repositoryDescriptor;
	
	protected RemoteRepository remoteRepository = null;
	
	public ArtifactRepositoryImpl() {
		// TODO Auto-generated constructor stub
	}

	public ArtifactRepositoryImpl(RepositoryDescriptor descriptor) {
		this();
		this.repositoryDescriptor = descriptor;
	}
	
	public void init() {
		Repository caveRepo = repositoryService.repository(repositoryDescriptor.getName());
		if (caveRepo == null) {
			caveRepo = createRepo();
		} else {
			caveRepo = checkUpdateRepo(caveRepo);
		}
	}
	
	protected Repository createRepo(){
		Repository repo = null;
		try {
			repo = repositoryService.create(
					repositoryDescriptor.getName(), 
					repositoryDescriptor.getLocation(), 
					Strings.isNullOrEmpty(repositoryDescriptor.getUrl())
						?"/repositories/" + repositoryDescriptor.getName()
						:repositoryDescriptor.getUrl(), 
					repositoryDescriptor.isProxy() && !Strings.isNullOrEmpty(repositoryDescriptor.getRemotes())
						?repositoryDescriptor.getRemotes()
						:null, 
					/*
					repositoryDescriptor.isProxy() && !Strings.isNullOrEmpty(repositoryDescriptor.getRemotes())
							&& !Strings.isNullOrEmpty(repositoryDescriptor.getLocation())
						?repositoryDescriptor.isMirror()
						:false*/
					repositoryDescriptor.isMirror(), 
					"karaf", 
					null, 
					null, 
					null, 
					null, 
					8);
			this.isHealthy = true;
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			this.isHealthy = false;
		}
		return repo;
	}
	
	protected Repository checkUpdateRepo(Repository repo) {
		try {
			log.info("repositoryDescriptor.getLocation(): [{}]; repo.getLocation(): [{}]", repositoryDescriptor.getLocation(), repo.getLocation());
			if (!Strings.isNullOrEmpty(repositoryDescriptor.getLocation())
					&& !repo.getLocation().equalsIgnoreCase(repositoryDescriptor.getLocation()))
				repositoryService.changeLocation(repositoryDescriptor.getName(), repositoryDescriptor.getLocation());
			if (!repo.getUrl().equalsIgnoreCase(repositoryDescriptor.getUrl()) 
					&& !Strings.isNullOrEmpty(repositoryDescriptor.getUrl()))
				repositoryService.changeUrl(repositoryDescriptor.getName(), repositoryDescriptor.getUrl());
			String repoProxy = repo.getProxy();
			boolean repoIsProxy = !Strings.isNullOrEmpty(repoProxy);
			if (repoIsProxy != repositoryDescriptor.isProxy()) {
				if (!repositoryDescriptor.isProxy)
					repositoryService.changeProxy(repositoryDescriptor.getName(), "", false);
				else
					repositoryService.changeProxy(repositoryDescriptor.getName()
							, repositoryDescriptor.getRemotes()
							, repositoryDescriptor.isMirror());
			}
			this.isHealthy = true;
		} catch (Exception e) {
			e.printStackTrace();
			this.isHealthy = false;
		}
		return repo;
	}
	
	@Override
	public DeployedArtifact deploy(Artifact artifact) {
		if (isHealthy() && isEnabled() && !isReadOnly()) {
			return deployArtifact(artifact);
		}
		return new DeployedArtifactImpl(artifact);
	}

	protected DeployedArtifact deployArtifact(Artifact artifact) {
		String location = "mvn:";
		
		if (!repositoryDescriptor.isProxy() || repositoryDescriptor.isMirror())
			deployToCave(artifact);
		
		if (repositoryDescriptor.isProxy())
			deployToRemoteMaven(artifact);
		
		return new DeployedArtifactImpl(artifact, location + artifact.getCoordinates().toString());
	}
	
	protected void deployToCave(Artifact artifact) {
		if (artifact.getContent() != null) {
			try {
				ArtifactCoordinates coords = artifact.getCoordinates();
				File tmpFile = File.createTempFile(getName()+"-", Calendar.getInstance().getTimeInMillis()+"");
				artifact.toFile(tmpFile);
				log.info("Artifact [{}] is written to [{}]", artifact.getCoordinates(), tmpFile.toURI().toURL().toString());
				repositoryService.addArtifact(tmpFile.toURI().toURL().toString()
						, coords.getGroupId()
						, coords.getArtifactId()
						, coords.getVersion()
						, coords.getType()
						, coords.getClassifier()
						, getName());	
				log.info("Artifact [{}] is deployed to [{}]", artifact.getCoordinates(), getName());
//				Files.deleteIfExists(tmpFil)
			} catch (IOException e) {
				log.error("deployToCave", e);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				log.error("deployToCave", e);
			}
		}
	}

	protected void deployToRemoteMaven(Artifact artifact) {
		try {
			initRemoteRepository();
			this.remoteRepository.deploy(artifact);
		} catch (Exception e) {
			log.error("deployToRemoteMaven", e);
		}
	}
	
	protected void initRemoteRepository() throws NoSuchMethodException, SecurityException, InstantiationException, IllegalAccessException, IllegalArgumentException, InvocationTargetException, MalformedURLException {
		if (this.remoteRepository == null) {
			this.remoteRepository = RemoteRepositoryFactory.create(new RemoteRepositoryDescriptor(repositoryDescriptor));
		}
	}
	
	@Override
	public String getName() {
		return repositoryDescriptor.getName();
	}

	@Override
	public boolean isHealthy() {
		return this.isHealthy;
	}

	@Override
	public boolean isEnabled() {
		return repositoryDescriptor.isEnabled();
	}

	@Override
	public boolean isReadOnly() {
		return repositoryDescriptor.isReadOnly();
	}
	
	@Override
	public boolean isSnapshotsAllowed() {
		return repositoryDescriptor.isSnapshots();
	}
	
	@Override
	public String getUrl() {
		return repositoryService.repository(getName()).getUrl();
	}

	public void setRepositoryService(RepositoryService repositoryService) {
		this.repositoryService = repositoryService;
	}
	
}
