/*-
 * ~~~~~~licensing~~~~~~
 * basic-auth-impl
 * ==========
 * Copyright (C) 2020 - 2021 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.esb.system.auth.basic.jpa.impl;

import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import ru.entaxy.esb.system.auth.basic.jpa.api.BasicAuthService;
import ru.entaxy.esb.system.auth.basic.jpa.api.entity.BasicAuthAccount;
import ru.entaxy.esb.system.auth.basic.jpa.api.entity.field.EncryptionAlgorithm;
import ru.entaxy.esb.system.auth.basic.jpa.api.exception.BadPasswordException;
import ru.entaxy.esb.system.auth.basic.jpa.util.EncryptionHelper;
import ru.entaxy.esb.system.core.permission.common.PermissionConstants;
import ru.entaxy.esb.system.core.permission.jpa.PermissionService;

import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Root;
import java.security.NoSuchAlgorithmException;
import java.util.Date;
import java.util.List;
import java.util.Optional;

public class BasicAuthServiceImpl implements BasicAuthService {

    private static final Logger LOG = LoggerFactory.getLogger(BasicAuthServiceImpl.class);

    private SessionFactory sessionFactory;
    private String еncryptionAlgorithm;
    private String encryptionSalt;
    private PermissionService permissionService;

    public void setSessionFactory(SessionFactory sessionFactory) {
        this.sessionFactory = sessionFactory;
    }

    public Session getSession() {
        return sessionFactory.openSession();
    }

    public void setЕncryptionAlgorithm(String еncryptionAlgorithm) {
        this.еncryptionAlgorithm = еncryptionAlgorithm;
    }

    public void setEncryptionSalt(String encryptionSalt) {
        this.encryptionSalt = encryptionSalt;
    }

    public PermissionService getPermissionService() {
        return permissionService;
    }

    public void setPermissionService(PermissionService permissionService) {
        this.permissionService = permissionService;
    }

    @Override
    public List<BasicAuthAccount> list() {
        List<BasicAuthAccount> list;
        try (Session s = getSession()) {
            s.getTransaction().begin();
            CriteriaQuery<BasicAuthAccount> cq = s.getCriteriaBuilder().createQuery(BasicAuthAccount.class);
            cq.from(BasicAuthAccount.class);
            list = s.createQuery(cq).getResultList();
            s.getTransaction().commit();
            s.close();
        }

        return list;
    }

    @Override
    public Optional<BasicAuthAccount> get(String login) {
        Optional<BasicAuthAccount> basicAuthAccount;
        try (Session s = getSession()) {
            s.getTransaction().begin();

            CriteriaBuilder builder = s.getCriteriaBuilder();
            CriteriaQuery<BasicAuthAccount> criteriaQuery = builder.createQuery(BasicAuthAccount.class);
            Root<BasicAuthAccount> root = criteriaQuery.from(BasicAuthAccount.class);
            criteriaQuery.where(builder.equal(root.get("login"), login));
            basicAuthAccount = s.createQuery(criteriaQuery).uniqueResultOptional();

            s.getTransaction().commit();
            s.close();
        }
        return basicAuthAccount;
    }

    public Optional<BasicAuthAccount> getBySystem(String systemUuid){
        Optional<BasicAuthAccount> basicAuthAccount;
        try (Session s = getSession()) {
            s.getTransaction().begin();

            CriteriaBuilder builder = s.getCriteriaBuilder();
            CriteriaQuery<BasicAuthAccount> criteriaQuery = builder.createQuery(BasicAuthAccount.class);
            Root<BasicAuthAccount> root = criteriaQuery.from(BasicAuthAccount.class);
            criteriaQuery.where(builder.equal(root.get("systemUUID"), systemUuid));
            basicAuthAccount = s.createQuery(criteriaQuery).uniqueResultOptional();

            s.getTransaction().commit();
            s.close();
        }
        return basicAuthAccount;
    };
    
    public Optional<BasicAuthAccount> get(Session session, String login) {
        Optional<BasicAuthAccount> basicAuthAccount;
        CriteriaBuilder builder = session.getCriteriaBuilder();
        CriteriaQuery<BasicAuthAccount> criteriaQuery = builder.createQuery(BasicAuthAccount.class);
        Root<BasicAuthAccount> root = criteriaQuery.from(BasicAuthAccount.class);
        criteriaQuery.where(builder.equal(root.get("login"), login));
        basicAuthAccount = session.createQuery(criteriaQuery).uniqueResultOptional();
        return basicAuthAccount;
    }

    @Override
    public Optional<BasicAuthAccount> getByAuthorizationHeaderHash(String authorizationHeaderHash) {
        Optional<BasicAuthAccount> basicAuthAccount;
        try (Session s = getSession()) {
            s.getTransaction().begin();

            CriteriaBuilder builder = s.getCriteriaBuilder();
            CriteriaQuery<BasicAuthAccount> criteriaQuery = builder.createQuery(BasicAuthAccount.class);
            Root<BasicAuthAccount> root = criteriaQuery.from(BasicAuthAccount.class);
            criteriaQuery.where(builder.equal(root.get("authorizationHeaderHash"), authorizationHeaderHash));
            basicAuthAccount = s.createQuery(criteriaQuery).uniqueResultOptional();

            s.getTransaction().commit();
            s.close();
        }
        return basicAuthAccount;
    }

    @Override
    public BasicAuthAccount save(BasicAuthAccount basicAuthAccount, boolean isExist) {
        try (Session s = getSession()) {
            s.getTransaction().begin();
            if (isExist) {
                s.update(basicAuthAccount);
            } else {
                s.persist(basicAuthAccount);
            }
            s.getTransaction().commit();
            s.close();
        }
        return basicAuthAccount;
    }

    @Override
    public BasicAuthAccount saveFull(String login, String passwordHash, String encryptionAlgorithm, String systemUUID,
                                     String authorizationHeaderHash, String description, String createdBy, String editedBy) {

        LOG.debug("Parameters " + login + " " + passwordHash + " " + encryptionAlgorithm
                + " " + systemUUID + " " + authorizationHeaderHash + " " + description + " " + createdBy + " " + editedBy);

        Optional<BasicAuthAccount> existingBasicAuthAccount = get(login);

        EncryptionAlgorithm encryptionAlgorithmObj = EncryptionAlgorithm.getByName(encryptionAlgorithm.toUpperCase());

        BasicAuthAccount basicAuthAccount = existingBasicAuthAccount.orElseGet(BasicAuthAccount::new);
        settingBasicAuthAccount(login, passwordHash, encryptionAlgorithmObj, systemUUID, authorizationHeaderHash,
                description, createdBy, editedBy, basicAuthAccount, existingBasicAuthAccount.isPresent());

        return save(basicAuthAccount, existingBasicAuthAccount.isPresent());
    }

    private BasicAuthAccount settingBasicAuthAccount(String login, String passwordHash, EncryptionAlgorithm encryptionAlgorithm,
                                                     String systemUUID, String authorizationHeaderHash,
                                                     String description, String createdBy, String editedBy,
                                                     BasicAuthAccount basicAuthAccount, boolean isExist) {
        if (!isExist) {
            basicAuthAccount.setLogin(login);
            basicAuthAccount.setCreateDate(new Date());
            basicAuthAccount.setCreatedBy(createdBy);
        } else {
            basicAuthAccount.setEditDate(new Date());
            basicAuthAccount.setEditedBy(editedBy);
        }

        basicAuthAccount.setPasswordHash(passwordHash);
        basicAuthAccount.setEncryptionAlgorithm(encryptionAlgorithm);
        basicAuthAccount.setSystemUUID(systemUUID);
        basicAuthAccount.setAuthorizationHeaderHash(authorizationHeaderHash);
        basicAuthAccount.setDescription(description);
        return basicAuthAccount;
    }

    @Override
    public BasicAuthAccount saveCommon(String login, String passwordHash, String systemUUID, String createdBy, String editedBy)
            throws NoSuchAlgorithmException, BadPasswordException {
        if (null != passwordHash && !passwordHash.isEmpty()) {
            passwordHash = EncryptionHelper.encrypt(passwordHash, this.еncryptionAlgorithm, this.encryptionSalt);
        } else {
            throw new BadPasswordException("Password not passed or empty!");
        }
        return saveFull(login, passwordHash, this.еncryptionAlgorithm, systemUUID, "", "", createdBy, editedBy);
    }

    @Override
    public BasicAuthAccount saveShort(String login, String passwordHash, String createdBy, String editedBy)
            throws NoSuchAlgorithmException, BadPasswordException {
        return saveCommon(login, passwordHash, "", createdBy, editedBy);
    }

    @Override
    public void remove(String login) {
        try (Session s = getSession()) {
            s.getTransaction().begin();

            Optional<BasicAuthAccount> basicAuthAccount = get(s, login);
            basicAuthAccount.ifPresent((basicAuthAccount1) -> removeProc(s, basicAuthAccount1));
            s.getTransaction().commit();
            s.close();
        }
    }

    private void removeProc(Session session, BasicAuthAccount account) {
        permissionService.removeAll(session, account.getId(), PermissionConstants.TYPE_ACCOUNT);
        session.flush();
        session.delete(BasicAuthAccount.class.getName(), account);
    }

    @Override
    public boolean check(String login, String password) {
        if (login != null && password != null) {
            Optional<BasicAuthAccount> basicAuthAccountOpt = get(login);
            if (basicAuthAccountOpt.isPresent()) {
                BasicAuthAccount basicAuthAccount = basicAuthAccountOpt.get();
                try {
                    String passedPasswordHash = EncryptionHelper.encrypt(password,
                            basicAuthAccount.getEncryptionAlgorithm().getAlgorithmName(),
                            this.encryptionSalt);
                    if (passedPasswordHash.equals(basicAuthAccount.getPasswordHash())) {
                        return true;
                    }
                } catch (NoSuchAlgorithmException e) {
                    LOG.error("Encription algorithm error", e);
                }
            }
        }
        return false;
    }


}
