/*-
 * ~~~~~~licensing~~~~~~
 * profile-commons
 * ==========
 * Copyright (C) 2020 - 2021 EmDev LLC
 * ==========
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.esb.system.profile.commons.profile_output;

import java.util.concurrent.CountDownLatch;

import org.apache.camel.Exchange;
import org.apache.camel.Header;
import org.apache.camel.ProducerTemplate;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.entaxy.esb.system.common.exception.ConnectorNotFound;
import ru.entaxy.esb.system.common.osgi.impl.CommonNamedReferenceListener;
import ru.entaxy.esb.system.profile.commons.connectors.out.OutConnector;

public class ProfileOutputImpl extends CommonNamedReferenceListener<OutConnector> implements ProfileOutput {

	private static final Logger log = LoggerFactory.getLogger(ProfileOutputImpl.class);
	
    protected String systemName;

    public void setSystemName(String systemName) {
        this.systemName = systemName;
    }

    @Override
    public OutConnector getOutConnector(String endpointName) {
        if (endpointName == null || registeredReferences.get(endpointName) == null) {
            for (OutConnector outConnector : registeredReferences.values()) {
                return outConnector;
            }
        }
        return registeredReferences.get(endpointName);
    }

    @Override
    public void send(Exchange exchange) {
    	/*
    	 * !!! WE ARE NOT SURE THE SENDING FROM PROCESSOR (COMPONENT) IS ALLOWED IN CAMEL !!!
    	 * but we managed to do this
    	 * 
    	 * we need another Thread to execute because when sending
    	 * from system to itself the execution stops in waiting forever
    	 * due to existing running ReactiveExecutor will be found for this thread
    	 * and the task will be placed in it's queue while being in
    	 * execution of the previous task from another Workers's queue
    	 * of the same ReactiveExecutor   
    	 * 
    	 * Example (without separate thread): 
    	 * 
    	 * profile_s1 (ExtendedCamelContext_1 -> ReactiveExecutor_1) 
    	 * -> task (we're in it now!) -> [switch to ExtendedCamelContext_2] producer (vmdirect) 
    	 * -> ExtendedCamelContext_2 (with ReactiveExecutor_2) (route_s1)
    	 * -> consumer (profile_s1) 
    	 * -> task {
    	 * 		Who will execute? Let's check if there's already a ReactiveExecutor !!for current Thread!!
    	 * 		in target context. If no, our executor will be used.
    	 * 		Oh! We've found ReactiveExecutor_1, it will execute. 
    	 * 	} -> enqueue (ReactiveExecutor_1)
    	 * -> (execution is still in ExtendedCamelContext_2 - route_s1)
    	 * -> wait for task to be executed.
    	 * 
    	 * Result: being in previous (current) task we're waiting for the next task 
    	 * in the same ReactiveExecutor's queue to be executed. 
    	 * This will never happen.
    	 *   
    	 */
    	CountDownLatch latch = new CountDownLatch(1);
    	Thread exec = new Thread(new Runnable() {
			
			@Override
			public void run() { 
		        ProducerTemplate template = exchange.getContext().createProducerTemplate();
		        template.send("direct-vm:profile-" + systemName + "-enter-point?block=true&timeout=60000", exchange);
		        latch.countDown();
			}
		});
    	exec.setContextClassLoader(Thread.currentThread().getContextClassLoader());
    	exec.start();
    	try {
    		// we have to wait the thread to be executed 
    		// because we're calling synchronously
			latch.await();
		} catch (InterruptedException e) {
			// e.printStackTrace();
			log.error("Error awaiting latch", e);
		} 
    }

    @Override
    public void sendTo(@Header("ENTAXY_EndpointName") String endpointName, Exchange exchange) {
        try {
            getOutConnector(endpointName).send(exchange);
        } catch (NullPointerException ex) {
            throw new ConnectorNotFound("Out connector for " + systemName + " not found");
        }
    }

    @Override
    protected String getObjectName(OutConnector outConnector) {
        return outConnector == null ? null : outConnector.getEndpointName();
    }
}
